/* Programa BAGGING que calcula el porcentaje de error, dicho programa saca muestras con    *
 *reemplazamiento para obetener los estimados  y luego clasificar para obtener los errores  */

#include <mpi.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#define CTE 2.506631   /*CTE=sqrt(2.0*PI)*/
#define WORK 0
#define DONE 1
#define RESULT 2
 
/*funcion que dado una matriz X, obtiene la n-esima division de X,  se debe ingresar un entero *
 * minimo, el # de filas de la division, y el # de columnas de la matriz                       */ 
 
 double ** test(double **X,int min,int n,int p){
          double **temp;
          int i;
          temp=calloc(n,sizeof(double *)); 
          for(i=0;i<n;i++)
              temp[i]= X[i+min]; 
          return temp;
          }

/*funcion que dado una matriz X, obtiene una muestra con reemplazamiento de las filas de X distintas de *
 * de las filas  que se uso para test, se debe ingresar un entero  minimo, maximo, # de filas de la     *
 * division  y el # de columnas de la matriz                       				        */ 
 
double ** training(double **X,int min,int max,int n,int p){
          double **temp;
          int i;
          temp=calloc(n,sizeof(double *));
          for(i=0;i<min;i++)
                temp[i]= X[rand()%min];  
          for(i=min;i<n-(max-min);i++)
                temp[i]= X[rand()%(n-max)+max]; 
           return temp;				
         }

/*funcion que genera numeros aleatorios desde cero hasta n*/

int * randperm(int n ){
       int i,j,min;
       int * a=NULL;
       int * indice=NULL;
       indice=(int *)calloc(n,sizeof(int));
       a=(int *)calloc(n,sizeof(int));   
       srand(time(NULL)); 
       for(i=0;i<n;i++)
           a[i]=rand()%n; 
       for ( i = 0; i < n;i++ ){
           min=i;
           for (j = 0; j < n; j++ )
                if ( a[j] < a[min] ) {
                    min=j;
                }
           a[min]=n;
           indice[i]=min; 
       }
       return indice;
       }

/*funcion que calcula el kernel gaussiano*/
double kernel(double x){
       return 1.0/(exp((x*x)/2.0))*CTE;        /*CTE =sqrt(2.0*PI)*/
       }

/*funcion que calcula el vector  del ancho de banda para una clase X, se debe ingresar el # de filas de la clase asi como  el # de columanas*/

double * banda(double **X, int n,int p){
       double s,d, Xmean,*h=NULL;
       int i,v;

       h=calloc(n,sizeof(double)); 
       for(v=0;v<p ;v++){
            Xmean=0;
            d=0;
            s=0;
            for(i=0;i<n;i++){
                 s+=X[i][v];
            }
            Xmean =s/n;       
            for(i=0;i<n;i++){
                d+=(X[i][v]-Xmean)*(X[i][v]-Xmean);
            }
            d=sqrt(d/(n-1));    
                                
            h[v]=d*(pow(4.0,1.0/(double)(p+4))/pow((double)((p+2)*n),1.0/(double)(p+4)));  

       }
       return h;
       }

/*funcion que calcula la suma de la productoria de los kernel dado una clase X y un test x, se debe ingresar el vector del *
 *ancho de banda, la iesima fila del test(i), el # de columnas(p) y el # de filas de la clase(m)                           */

 double suma_productoria(double **x,double *h, double **X,int i, int p,int m ){
        double s=0.0,k,t=0.0;
        int j,v;
        for(j=0;j<m ;j++){
                k=1.0;
                for(v=0;v<p;v++){
                     t=kernel((x[i][v]-X[j][v])/h[v]);
                     k=k*t;
                }
              s=s+k; 
        }
        return s;
        }

 /*funcion que calcula el estimado del test x dado una clase X , se debe ingresar el vector del ancho de banda (h), el # de filas del test(n),*
  *el # de columnas(p), el # de filas para una clase del training(m)  y el numero de filas de la matriz original(N)                           */
 
 double * estimado(double **x,double *h,double **X,int n, int p,int m,int N){
       int v,i;
       double s=0.0,H=1.0;
       double *f=NULL;
       double pro=(double)m/(double)(N-n);

       f=calloc(n,sizeof(double));    
       for(v=0;v<p;v++)
                   H=H*h[v];          
       H=H*(double)m;
       for (i=0;i<n;i++){
              s=suma_productoria(x,h,X,i,p,m);
              f[i]=pro*(s/H);         
       }
       return f;
       }
/* funcion que calcula los errores dado una matriz M de dimension(n_boos X n_test), se debe ingresar el vector de indices de la muestra original(k)  *
 * el # de columnas de M(n_test), el # de filas de M(n_boos), el # de clases(m) y un entero minimo(min) para relizar las comparaciones con el vector k */

 int error(int **M,int * k_indices ,int n_test,int n_boos,int m,int min ){
       int i,j,e=0,k,max;
       int *clase=NULL;
       clase=calloc(m+1,sizeof(int)); 
       for(j=0;j<n_test;j++){
             for(i=0;i<m+1;i++) clase[i]=0;  
             for(i=0;i<n_boos;i++) clase[M[i][j]]++; 
             max=0;
             k=0;            
             for(i=0;i<m+1;i++){
                 if(clase[i]>max){   
                      max=clase[i];
                      k=i;      
                 }
                      
             } 
             if(k_indices[j+min]!=k) 
                 e++;
       }
       return e;
       }

/*funcion que calcula el vector de maximos dado una matriz f de estimados, se debe ingresar el # de filas de f (m), el # de columnas(n)*/
int  * maximos(double **f,int n,int m ){
      int i,j,k;
      int *max=NULL;
      double maximo;
      max = calloc(n,sizeof(int)); 
      for(j=0;j<n;j++){
             maximo=0;
             for(i=0;i<m;i++){
                   if(f[i][j]>=maximo){
                        maximo=f[i][j];
                        k=i+1;
                    }
             }
             max[j]=k;     
       }
       return max;
       }

/* funcion que sirve para imprimir matrices de tipo double en la pantalla, se ingresar la matriz con sus respectivas filas y columnas*/
void print(double **X,int n,int m){
       int i,j;
        for(i=0;i<n;i++){
                for(j=0;j<m;j++)
                      printf("%e\t",X[i][j]);
                 printf("\n");
           }

       }
       
/* funcion que sirve para imprimir matrices de tipo entero  en la pantalla, se ingresar la matriz con sus respectivas filas y columnas*/
void print2(int  **X,int n,int m){
       int i,j;
        for(i=0;i<n;i++){
                for(j=0;j<m;j++)
                      printf("%d\t",X[i][j]);
                 printf("\n");
           }

       }

/*funcion que calcula la matriz de estimados, se le debe ingresar la matriz test(x), matriz training(X) , # de filas del training(n),  *
 * # de columnas(p), # de clases(m), # de filas del test(r), y el # de filas de la matriz original (N)                                 */

double **estimated(double ** x, double ** X,int n, int p,int m,int r,int N){
       double **f;
       double ** temp;
       double *h=NULL;
       int i,j,cont;

       f=calloc(m,sizeof(double *));   
       temp=calloc(n,sizeof(double *));
       h=calloc(p,sizeof(double));     
       for(j=0;j<m;j++){
            cont=0;
            for(i=0;i<n;i++)
                    if(X[i][p-1]==(double)(j+1)){
                       temp[cont++]= X[i];	
                    }

            h=banda(temp,cont,p-1);        
            f[j] = estimado(x,h,temp,r,p-1,cont,N); 

      }
      return f;
}



int *max_estimado(double ** x, double ** X,int n, int p,int m,int r){
       
       int i,j,k,cont,n_cross= n-r;
       int *max=NULL;
       double maximo; 
       double **f;
       double ** temp;
       double *h=NULL;

       f=calloc(m,sizeof(double *));   
       temp=calloc(n_cross,sizeof(double *));
       h=calloc(p,sizeof(double));     
       for(j=0;j<m;j++){
            cont=0;
            for(i=0;i<n_cross;i++)
                    if(X[i][p-1]==(double)(j+1)){
                       temp[cont++]= X[i];	
                    }

            h=banda(temp,cont,p-1);        
            f[j] = estimado(x,h,temp,r,p-1,cont,n); 

      }
                  
      max = calloc(r,sizeof(int)); 
      for(j=0;j<r;j++){
             maximo=0;
             for(i=0;i<m;i++){
                   if(f[i][j]>=maximo){
                        maximo=f[i][j];
                        k=i+1;
                    }
             }
             max[j]=k;     
       }
      
      return max;
      
}




/*funcion que calcula los errore al clasificar, se debe ingresar el vector de los datos originales(x), el numero de filas(n), el numero de *
 * muestras boostrapeadas (n_boos), el # de clases(m) y el # de columnas (p)								   */

int boostrap(double * x,int n,int n_boos,int m,int p){
    int i,j,r=n/10,e=0,err=0,b;
    int *indice=NULL;
    double **f;
    double **tes;
    double **X;
    double **trai;
    int **M;
    int *k=NULL;
    M=calloc(n_boos,sizeof(int *));   
    X=calloc(n,sizeof(double *));     
    tes=calloc(n,sizeof(double *));   
    trai=calloc(n,sizeof(double *));  
    k=calloc(n,sizeof(int));          
    indice=(int *)calloc(n,sizeof(int));
    
    indice=randperm(n);  
    for(i=0;i<n;i++){
          X[i]=x+indice[i]*p; 
          k[i]=X[i][p-1]; 
                    
    }
    j=0;
    if(n%10==0){	
          for(i=0;i<10;i++){
             j=j+r;
             tes=test(X,j-r,r,p);    
             for(b=0;b<n_boos;b++){
                  trai=training(X,j-r,j,n,p);    
                  /*f=estimated(tes,trai,n-r,p,m,r,n); 
                  M[b]=maximos(f,r,m); */        
		  M[b]=max_estimado(tes,trai,n,p,m,r);        
             }

             e=error(M,k,r,n_boos,m,j-r);      
             err+=e;			       

          }
     }
     else {			
           for(i=0;i<n%10;i++){
              j=j+r+1;
              tes=test(X,j-r-1,r+1,p);                 
              for(b=0;b<n_boos;b++){
                  trai=training(X,j-r-1,j,n,p);        
		  /*                 f=estimated(tes,trai,n-(r+1),p,m,r+1,n); 
                  M[b]=maximos(f,r+1,m);*/          
		  M[b]=max_estimado(tes,trai,n,p,m,r+1);        
             }
             e=error(M,k,r+1,n_boos,m,j-(r+1));   
             err+=e;                              
                       
           }
           for(i=n%10;i<10;i++){
               j=j+r;
               tes=test(X,j-r,r,p);               
               for(b=0;b<n_boos;b++){
                  trai=training(X,j-r,j,n,p);/*       
                  f=estimated(tes,trai,n-r,p,m,r,n);
                  M[b]=maximos(f,r,m);       */
		  M[b]=max_estimado(tes,trai,n,p,m,r);           
               } 
               e=error(M,k,r,n_boos,m,j-r);     
               err+=e;				
           }
     }
     return err;
     }

/* funcion main, se debe ingresar en la linea de comandos lo siguiente:  *
 * argv [1] la matriz original; argv [2] el numero de filas, argv [3]   *
 * argv [3] el numero de columnas; argv [4] el numero de clases y      *
 * argv [5] el numero de muestras boostrapeadas                       */

int main(int argc, char * argv[]){

    FILE *in;      
    int n_boos,parametro[4];
    int n,m,p,rank,size;
    double *x=NULL;
    double tiempo;
    MPI_Status status;


  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  

		/*proceso maestro  */

if(rank==0)
{
    int r,t,b,i,j=0,s,err=0,e;
    int *indice=NULL;
    double **f;
    double **tes;
    double **X;
    double **trai;
    int **M;
    int *max;
    int *k=NULL;
   
 if(argc!=6){
       fprintf(stderr,"Uso de %s\n  nombre_del_archivo #_de_filas  #_de_columnas #_de_clases #_de_muestras_boostrapeadas\n",argv[0]);
       exit(1);
    }
   
    tiempo= MPI_Wtime();

    srand(time(0x0));   


    n=atoi(argv[2]);    
    p=atoi(argv[3]);    
    m=atoi(argv[4]);    
    n_boos=atoi(argv[5]);    


    
    parametro[0]=n; 
    parametro[1]=p;
    parametro[2]=m;
    parametro[3]=n_boos;
   
 
    MPI_Bcast(parametro,4,MPI_INT,0,MPI_COMM_WORLD);

    
    indice=calloc(n,sizeof(int));
    k=calloc(n,sizeof(int));
    x=calloc(n*p,sizeof(double));    
    tes=calloc(n,sizeof(double *)); 
    trai=calloc(n,sizeof(double *)); 
    f=calloc(m,sizeof(double *)); 
    X=calloc(n,sizeof(double *)); 
    M=calloc(n_boos,sizeof(int *));   
    max = (int *)calloc(n/10+1,sizeof(int));

   
    in=fopen(argv[1],"r");        
    if(in==NULL){                 
         printf("problemas con el archivo de salida");
         exit(1);
    }


    while (!feof(in)){       
             fscanf(in,"%lf",&x[j]); 
             j++;
      }
   fclose(in); 

   if(size==1){
   err= boostrap(x, n, n_boos, m, p);
 
   printf("\nERRORES %d \n",err);
   printf("\nPORCENTAJE DE ERRORES %f \n",(double)err/(double)n);
   tiempo=MPI_Wtime()-tiempo;
   printf("\nEl tiempo paralelo es %lf \n: ",tiempo);
   MPI_Finalize(); 
   return 0;

   }

   else{

   err=0;
   MPI_Bcast(x,n*p,MPI_DOUBLE,0,MPI_COMM_WORLD);
  
   indice=randperm(n); 
   MPI_Bcast(indice,n,MPI_INT,0,MPI_COMM_WORLD);
    
    for(i=0;i<n;i++){
          X[i]=x+indice[i]*p; 
	  /*  X[i]=x+i*p; se hace una copia del vector x a la matriz X*/
          k[i]=X[i][p-1];   
    }

    r=n/10;
    j=0;

    if(n%10==0){	
          for(s=0;s<10;s++){
              b=0;
	     j=j+r;

              for(i=0;i<size-1;i++) {
                  MPI_Ssend(&t,0,MPI_INT,i+1,WORK,MPI_COMM_WORLD);
               }
              for(i=size-1;i<n_boos;i++)  {
             
		   MPI_Recv(max,r,MPI_INT,MPI_ANY_SOURCE,RESULT,MPI_COMM_WORLD,&status);
                   M[b++]= max ;
		   /*	   printf("\niter   %d Recibi  %d  MASTER  ERROR  %d\n",s,status.MPI_SOURCE,e);*/                         
		   MPI_Ssend(&t,0,MPI_INT,status.MPI_SOURCE,WORK,MPI_COMM_WORLD);    
                  
               }   

               for(i=0;i<size-1;i++) {
                    MPI_Recv(max,r,MPI_INT,MPI_ANY_SOURCE,RESULT,MPI_COMM_WORLD,&status);
                    M[b++]=max;
		    /*		    printf("\niter   %d Recibi  %d  MASTER  ERROR  %d\n",s,status.MPI_SOURCE,e);*/                         
		    MPI_Ssend(&t,0,MPI_INT,status.MPI_SOURCE,DONE,MPI_COMM_WORLD);

               }
	                  
               e=error(M,k,r,n_boos,m,j-r);      
               err+=e;			       
	       /*	       printf("\niter   %d   MASTER  ERROR  %d\n",s,e);*/                         

          }



     }
     else {			
           for(s=0;s<n%10;s++){
              j=j+r+1;
	      b=0;
              for(i=0;i<size-1;i++) {
                  MPI_Ssend(&t,0,MPI_INT,i+1,WORK,MPI_COMM_WORLD);
               }
              for(i=size-1;i<n_boos;i++)  {
             
		   MPI_Recv(max,r+1,MPI_INT,MPI_ANY_SOURCE,RESULT,MPI_COMM_WORLD,&status);
                   M[b++]= max ;
		   /*		   printf("\niter   %d Recibi  %d  MASTER  ERROR  %d\n",s,status.MPI_SOURCE,e);*/                         
		   MPI_Ssend(&t,0,MPI_INT,status.MPI_SOURCE,WORK,MPI_COMM_WORLD);    
                  
               }   

               for(i=0;i<size-1;i++) {
                    MPI_Recv(max,r+1,MPI_INT,MPI_ANY_SOURCE,RESULT,MPI_COMM_WORLD,&status);
                    M[b++]=max;
		    /*		    printf("\niter   %d Recibi  %d  MASTER  ERROR  %d\n",s,status.MPI_SOURCE,e);*/                         
                    MPI_Ssend(&t,0,MPI_INT,status.MPI_SOURCE,DONE,MPI_COMM_WORLD);

               }
               e=error(M,k,r+1,n_boos,m,j-(r+1));      
               err+=e;			       
	       /*      printf("\niter   %d   MASTER  ERROR  %d\n",s,e); */                        
           }

          
           for(s=n%10;s<10;s++){
               j=j+r;
	       b=0;
              for(i=0;i<size-1;i++) {
                  MPI_Ssend(&t,0,MPI_INT,i+1,WORK,MPI_COMM_WORLD);
               }
              for(i=size-1;i<n_boos;i++)  {
             
		   MPI_Recv(max,r,MPI_INT,MPI_ANY_SOURCE,RESULT,MPI_COMM_WORLD,&status);
                   M[b++]= max ;
		   /*		   printf("\niter   %d Recibi  %d  MASTER  ERROR  %d\n",s,status.MPI_SOURCE,e);*/                         
		   MPI_Ssend(&t,0,MPI_INT,status.MPI_SOURCE,WORK,MPI_COMM_WORLD);    
                  
               }   

               for(i=0;i<size-1;i++) {
                    MPI_Recv(max,r,MPI_INT,MPI_ANY_SOURCE,RESULT,MPI_COMM_WORLD,&status);
                    M[b++]=max;
		    /*		    printf("\niter   %d Recibi  %d  MASTER  ERROR  %d\n",s,status.MPI_SOURCE,e); */                        
                    MPI_Ssend(&t,0,MPI_INT,status.MPI_SOURCE,DONE,MPI_COMM_WORLD);

               }
	       e=error(M,k,r,n_boos,m,j-r);     
               err+=e;				
	       /*  printf("\niter   %d   MASTER  ERROR  %d\n",s,e);*/                         
           }
     }

tiempo=MPI_Wtime()-tiempo;

printf("\nDATA%s PROCRS  %d  N_BOOST %d PORC DE ERRORES %f tiempo paralelo es %.3lf \n: ",argv[1],size,n_boos,(double)err/(double)n,tiempo);
 MPI_Finalize(); 


 }

}

 else     /* SLAVE */
 
 { 

    int t,i,j,r,e=0,err=0,b,p,cont=0;
    int *indice=NULL;
    double *x;
    double **f;
    double **tes;
    double **X;
    double **trai;
    int *M ,parametro[4];
    int *k=NULL;
    MPI_Status status;   

    if(size==1){
     MPI_Finalize(); 
     return 0;
    }
    else{
  
    MPI_Bcast(parametro,4,MPI_INT,0,MPI_COMM_WORLD);
    x=calloc(parametro[0]*parametro[1],sizeof(double));      
    X=calloc(parametro[0],sizeof(double *));     
    tes=calloc(parametro[0],sizeof(double *));   
    trai=calloc(parametro[0],sizeof(double *));  
    k=calloc(parametro[0],sizeof(int));          
    indice=(int *)calloc(parametro[0],sizeof(int));


   MPI_Bcast(x,parametro[0]*parametro[1],MPI_DOUBLE,0,MPI_COMM_WORLD);

   MPI_Bcast(indice,parametro[0],MPI_INT,0,MPI_COMM_WORLD);

    
    for(i=0;i<parametro[0];i++){
          X[i]=x+indice[i]*parametro[1]; 
	  /*  X[i]=x+i*p; se hace una copia del vector x a la matriz X*/
          k[i]=X[i][parametro[1]-1];   
    }

 j=0;
 r=parametro[0]/10;


 if(parametro[0]%10==0){

     for(i=0;i<10;i++){
            j=j+r;
            tes=test(X,j-r,r,parametro[1]);
	    /*  M=(int *)calloc(r,sizeof(int));*/
	    cont=0;       
            while(1) {
	      cont++;
                MPI_Recv(&t,0,MPI_INT,0,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
                if(status.MPI_TAG==WORK) {   		
		  trai=training(X,j-r,j,parametro[0],parametro[1]);    
		  M=max_estimado(tes,trai,parametro[0],parametro[1],parametro[2],r);        
      	          MPI_Ssend(M,r,MPI_INT,0,RESULT,MPI_COMM_WORLD);

                 }

               else
                 break;
             } 
	    /* printf("\niteracion %d  contador %d SLAVE %d\n",i,cont-1,rank);*/
     }
  }
 else{
 
     for(i=0;i<parametro[0]%10;i++){
        j=j+r+1;
            tes=test(X,j-(r+1),r+1,parametro[1]);
	    /*  M=(int *)calloc(r,sizeof(int));*/       
	    cont=0;
            while(1) {
	      cont++;
                MPI_Recv(&t,0,MPI_INT,0,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
                if(status.MPI_TAG==WORK) {   		
		  trai=training(X,j-(r+1),j,parametro[0],parametro[1]);    
		  M=max_estimado(tes,trai,parametro[0],parametro[1],parametro[2],r+1);        
      	          MPI_Ssend(M,r+1,MPI_INT,0,RESULT,MPI_COMM_WORLD);

                 }

               else
                 break;
             } 
	    /* printf("\niteracion %d  contador %d SLAVE %d\n",i,cont-1,rank);*/

     }
     
     for(i=parametro[0]%10;i<10;i++){
        j=j+r;
            tes=test(X,j-r,r,parametro[1]);
	    /*  M=(int *)calloc(r,sizeof(int));*/       
	    cont=0;
            while(1) {
	      cont++;
                MPI_Recv(&t,0,MPI_INT,0,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
                if(status.MPI_TAG==WORK) {   		
		  trai=training(X,j-r,j,parametro[0],parametro[1]);    
		  M=max_estimado(tes,trai,parametro[0],parametro[1],parametro[2],r);        
		  MPI_Ssend(M,r,MPI_INT,0,RESULT,MPI_COMM_WORLD);

                 }

               else
                 break;
             } 
	    /* printf("\niteracion %d  contador %d SLAVE %d\n",i,cont-1,rank);*/
     }
          
 } 

 
 MPI_Finalize(); 
 }


 return 0;
}
}
 /* Compile this program
 [elozano@penguin1 elozano]$ gcc -o  time_sys time_sys.c -lm

and run it as follows
[elozano@penguin1 elozano]$ time -p ./time_sys
*/









