#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "mpi.h"
#define CTE  2.506631   /*CTE=sqrt(2.0*PI)*/

double ** test(double **X,int min,int n,int p){
   double **temp;
   int i;
   temp=calloc(n,sizeof(double *));
   for(i=0;i<n;i++)
     temp[i]= X[i+min];         
   return temp;
 }

double ** training(double **X,int min,int max,int n,int p){
  double **temp;
  int i,r,n_trai=n-(max-min);
  temp=calloc(n,sizeof(double *));  
  for(i=0;i<n_trai;i++){
    r=rand()%n_trai;  
    if(r<min)             
      temp[i]= X[r];
    else
      temp[i]= X[r+(max-min)];
  }
  return temp;
}

/*funcion que genera numeros aleatorios desde cero hasta n*/
int * randperm(int n ){
  int i,j,min;
  int * a=NULL;
  int * indice=NULL;
  indice=(int *)calloc(n,sizeof(int)); 
  a=(int *)calloc(n,sizeof(int));      
  srand(time(NULL));                   
  for(i=0;i<n;i++)
    a[i]=rand()%n;                   
  for ( i = 0; i < n;i++ ){
    min=i;
    for (j = 0; j < n; j++ )
      if ( a[j] < a[min] ) {
	min=j;
      }
    a[min]=n;
    indice[i]=min;                   
  }
  return indice;
}

/*funcion que calcula el kernel gaussiano*/

double kernel(double x){
  return 1.0/(exp((x*x)/2.0)*CTE);   /* CTE = sqrt(2.0*PI);*/
}

/*funcion que calcula el vector  del ancho de banda para una clase X, se debe ingresar el # de filas de la clase asi como  el # de columanas*/

double * banda(double **trai, int **clases,int n_trai,int j,int p){
  double s,d, Xmean;
  double *h=NULL;
  int i,v;
  h=calloc(p,sizeof(double));                     
  for(v=0;v<p ;v++){
    Xmean=0;
    d=0;
    s=0;
    for(i=0;i<n_trai;i++){
      s+=trai[clases[j][i]][v];
    }	 
    Xmean =s/n_trai;                              
    for(i=0;i<n_trai;i++){
      d+=(trai[clases[j][i]][v]-Xmean)*(trai[clases[j][i]][v]-Xmean);
    }
    d=sqrt(d/(n_trai-1));                         	 
    h[v]=d*(pow(4.0,1.0/(double)(p+4))/pow((double)((p+2)*n_trai),1.0/(double)(p+4)));
  }
  return h;
}


double *par_kernel_usual(double **x,double **trai,int **clases,double *h,int n_trai,int c,int n_cross, int p,int N,int rank ,int size){
  int v,j,i,residuo=n_cross-size*(n_cross/size),q=n_cross/size;
  int inicial,final;
  double t,k,s=0.0,H=1.0;
  double *f=NULL;
  double pro=(double)n_trai/(double)(N-n_cross);
  for(v=0;v<p;v++)
    H=H*h[v];
  H=H*(double)n_trai; 
  f=calloc(q+1,sizeof(double));
  if(rank<residuo){
    inicial=rank*(q+1);
    final=(rank+1)*(q+1);
  }
  else{
    inicial=residuo+rank*q;
    final=residuo+(rank+1)*q;
  }
  for (i=inicial;i<final;i++){
    s=0.0;                   
    for(j=0;j<n_trai ;j++){
      k=1.0;
      for(v=0;v<p;v++){
	t=kernel((x[i][v]-trai[clases[c][j]][v])/h[v]);
	k=k*t;
      }
      s=s+k;
    }
    f[i-inicial]=pro*(s/H); 
  }
  return f;
}


int par_estimado_kernel_usual(double **x, double **X,int *k_indices,int n, int p,int m,int n_cross,int j_val,int rank,int size){
 
  double e;
  double **f;
  double *h=NULL;
  int i,j,ntrai=n-n_cross,min=j_val-n_cross;
  int  error(double **,int *,int,int,int,int ,int); 
  int *n_trai=NULL;       
  int **clases;

  n_trai=calloc(m+1,sizeof(int));
  clases =calloc(m+1,sizeof(int *));
  h=calloc(p,sizeof(double));
  f=calloc(m,sizeof(double *)); 
  for(j=0;j<=m;j++){
    n_trai[j]=0; 
    clases[j]=calloc(n,sizeof(int));   
  }  
  for(i=0;i<ntrai;i++){
    clases[(int)X[i][p-1]][n_trai[(int)X[i][p-1]]++]=i;
  }
  for(j=0;j<m;j++){
    h=banda(X,clases,n_trai[j+1],j+1,p-1);                  
    f[j]=par_kernel_usual(x,X,clases,h,n_trai[j+1],j+1,n_cross,p-1,n,rank,size);
  }
  e=error(f,k_indices,n_cross,m,min,rank,size);
  return e; 
}

int error(double **f,int * k_indices ,int n_cross,int m,int min ,int rank,int size){
  int i,j,e=0,k,q=n_cross/size,residuo=n_cross-size*(n_cross/size),inicial;
  double max;
  if(rank<residuo){
    inicial=min+rank*(q+1);
    q=q+1;
  }
  else
    inicial=min+residuo+rank*q;
 
 for(j=0;j<q;j++){
    max=0;
    for(i=0;i<m;i++){
      if(f[i][j]>=max){
	max=f[i][j];
	k=i+1;
      }
    }            
    if(k_indices[j+inicial]!=k) 
      e++;
  }
  return e;
}

int main(int argc, char * argv[]){
  
  int rank,size;
  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  if(rank==0){
    FILE *in;   
    int n,m,p,r,t;
    int par_e=0,err=0,e=0;
    int parametro[4];
    int i,j=0,promedio;
    int *k;
    int *indice;
    double time;
    double **X;
    double **f;
    double **trai;
    double **tes;
    double *x=NULL;
    if(argc!=6){
      fprintf(stderr,"Uso de %s\n  nombre_del_archivo #_de_filas  #_de_columnas #_de_clases #_de_corridas_promedio\n",argv[0]);
      exit(1);
    }
    time=MPI_Wtime();
    n=atoi(argv[2]);           
    p=atoi(argv[3]);           
    m=atoi(argv[4]);           
    promedio = atoi(argv[5]);  
    parametro[0]=n; 
    parametro[1]=p;
    parametro[2]=m;
    parametro[3]=promedio;
    MPI_Bcast(parametro,4,MPI_INT,0,MPI_COMM_WORLD);
    indice=calloc(n,sizeof(int));
    k=calloc(n,sizeof(int));
    x=calloc(n*p,sizeof(double));    
    tes=calloc(n,sizeof(double *)); 
    trai=calloc(n,sizeof(double *)); 
    f=calloc(m,sizeof(double *)); 
    X=calloc(n,sizeof(double *)); 
   
    in=fopen(argv[1],"r");        
    if(in==NULL){                
         printf("problemas con el archivo de salida");
         exit(1);
    }
    while (!feof(in)){   
             fscanf(in,"%lf",&x[j]);
             j++;
      }
  fclose(in);  
  MPI_Bcast(x,n*p,MPI_DOUBLE,0,MPI_COMM_WORLD); 
  for(t=0;t<promedio;t++){
    indice=randperm(n); 
    MPI_Bcast(indice,n,MPI_INT,0,MPI_COMM_WORLD);
    for(i=0;i<n;i++){
      X[i]=x+indice[i]*p; 
      k[i]=X[i][p-1];   
    }
    r=n/10;
    j=0;     
    if(n%10==0){      
      for(i=0;i<10;i++){
	j=j+r;
	trai=training(X,j-r,j,n,p);  
	tes=test(X,j-r,r,p);         
	e=par_estimado_kernel_usual(tes,trai,k,n,p,m,r,j,rank,size);  
	MPI_Reduce(&e,&par_e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
	err+=par_e;                               
      }    
    }
    else {       
      for(i=0;i<n%10;i++){
	j=j+r+1;
	trai=training(X,j-r-1,j,n,p);     
	tes=test(X,j-r-1,r+1,p);           
	e=par_estimado_kernel_usual(tes,trai,k,n,p,m,r+1,j,rank,size);  
	MPI_Reduce(&e,&par_e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
	err+=par_e;                          
      }
      for(i=n%10;i<10;i++){
	j=j+r;
	trai=training(X,j-r,j,n,p);       
	tes=test(X,j-r,r,p);              
	e=par_estimado_kernel_usual(tes,trai,k,n,p,m,r,j,rank,size);  
	MPI_Reduce(&e,&par_e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
	err+=par_e;                             
      }
    }
  }
  time=MPI_Wtime()-time;

 printf("\nPROGRAMA %s DATA %s PROCRS  %d  ERRORES %lf tiempo paralelo es %.3lf \n: ",argv[0],argv[1],size,(double)err/(double)(n*promedio),(double)time/(double)promedio);
  MPI_Finalize();

  }
  
 else{
 
  int i,j,r,t,e=0,par_e=0,err=0,temp[4];
  int par_cross(double *,int,int,int,int,int);
  int *indice;
  int *k;
  double *x=NULL;
  double *h;
  double **f;
  double **trai;
  double **training(double **,int,int,int,int);
  double **tes;
  double **test(double **,int,int,int);
  double **X;
  MPI_Bcast(temp,4,MPI_INT,0,MPI_COMM_WORLD);
  indice=calloc(temp[0],sizeof(int));
  k=calloc(temp[0],sizeof(int));
  tes=calloc(temp[0],sizeof(double *)); 
  trai=calloc(temp[0],sizeof(double *)); 
  x=calloc(temp[0]*temp[1],sizeof(double));
  X=calloc(temp[0],sizeof(double *));
  f=calloc(temp[2],sizeof(double *));
  h=calloc(temp[1],sizeof(double));    
  MPI_Bcast(x,temp[0]*temp[1],MPI_DOUBLE,0,MPI_COMM_WORLD);
  for(t=0;t<temp[3];t++){
    MPI_Bcast(indice,temp[0],MPI_INT,0,MPI_COMM_WORLD);
    for(i=0;i<temp[0];i++){
      X[i]=x+indice[i]*temp[1]; 
      k[i]=X[i][temp[1]-1];   
    }
    r= temp[0]/10;
    j=0;
    if(temp[0]%10==0){      
      for(i=0;i<10;i++){
	j=j+r;
	trai=training(X,j-r,j,temp[0],temp[1]);  
	tes=test(X,j-r,r,temp[1]);         
	e=par_estimado_kernel_usual(tes,trai,k,temp[0],temp[1],temp[2],r,j,rank,size);  
	MPI_Reduce(&e,&par_e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
      }
    }
    else {       
      for(i=0;i<temp[0]%10;i++){
	j=j+r+1;
	trai=training(X,j-r-1,j,temp[0],temp[1]);
	tes=test(X,j-r-1,r+1,temp[1]);           
	e=par_estimado_kernel_usual(tes,trai,k,temp[0],temp[1],temp[2],r+1,j,rank,size);
	MPI_Reduce(&e,&par_e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);	err+=par_e;                          
      }
      for(i=temp[0]%10;i<10;i++){
	j=j+r;
	trai=training(X,j-r,j,temp[0],temp[1]);
	tes=test(X,j-r,r,temp[1]);              
	e=par_estimado_kernel_usual(tes,trai,k,temp[0],temp[1],temp[2],r,j,rank,size);  
	MPI_Reduce(&e,&par_e,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
      }
    }
  }
  MPI_Finalize();
 }
return 0;
}












